"""Constants for ExpandedTestCases package."""

# Third-Party Imports.
from django.conf import settings


# Imports that may not be accessible, depending on local python environment setup.
try:
    from colorama import Back, Fore, Style
    COLORAMA_PRESENT = True
except ImportError:
    COLORAMA_PRESENT = False


# Underline style definition for debug printing.
UNDERLINE = '\u001b[4m'
UNDERLINE_RESET = '\u001b[0m'


# General output/color format settings.
ETC_OUTPUT_ERROR_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_OUTPUT_ERROR_HEADER_COLOR',
    '{0}{1}{2}'.format(Fore.RED, Back.RESET, Style.NORMAL) if COLORAMA_PRESENT else '',
))
ETC_OUTPUT_EXPECTED_MATCH_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_OUTPUT_EXPECTED_MATCH_COLOR',
    '{0}{1}{2}'.format(Fore.CYAN, Back.RESET, Style.NORMAL) if COLORAMA_PRESENT else '',
))
ETC_OUTPUT_EXPECTED_ERROR_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_OUTPUT_EXPECTED_ERROR_COLOR',
    '{0}{1}{2}'.format(Fore.BLACK, Back.CYAN, Style.NORMAL) if COLORAMA_PRESENT else '',
))
ETC_OUTPUT_ACTUALS_MATCH_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_OUTPUT_ACTUALS_MATCH_COLOR',
    '{0}{1}{2}'.format(Fore.MAGENTA, Back.RESET, Style.NORMAL) if COLORAMA_PRESENT else '',
))
ETC_OUTPUT_ACTUALS_ERROR_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_OUTPUT_ACTUALS_ERROR_COLOR',
    '{0}{1}{2}'.format(Fore.BLACK, Back.MAGENTA, Style.NORMAL) if COLORAMA_PRESENT else '',
))
ETC_OUTPUT_EMPHASIS_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_OUTPUT_EMPHASIS_COLOR',
    (Style.BRIGHT if COLORAMA_PRESENT else '') + UNDERLINE,
))
ETC_OUTPUT_RESET_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_OUTPUT_RESET_COLOR',
    Style.RESET_ALL if COLORAMA_PRESENT else UNDERLINE_RESET,
))

# Output/color formatting for response sections.
ETC_RESPONSE_DEBUG_URL_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_URL_COLOR',
    Fore.YELLOW if COLORAMA_PRESENT else '',
))
ETC_RESPONSE_DEBUG_CONTENT_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_CONTENT_COLOR',
    Fore.WHITE if COLORAMA_PRESENT else '',
))
ETC_RESPONSE_DEBUG_HEADER_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_HEADER_COLOR',
    Fore.CYAN if COLORAMA_PRESENT else '',
))
ETC_RESPONSE_DEBUG_CONTEXT_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_CONTEXT_COLOR',
    Fore.BLUE if COLORAMA_PRESENT else '',
))
ETC_RESPONSE_DEBUG_SESSION_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_SESSION_COLOR',
    Fore.MAGENTA if COLORAMA_PRESENT else '',
))
ETC_RESPONSE_DEBUG_MESSAGE_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_MESSAGES_COLOR',
    Fore.CYAN if COLORAMA_PRESENT else '',
))
ETC_RESPONSE_DEBUG_FORM_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_FORMS_COLOR',
    Fore.BLUE if COLORAMA_PRESENT else '',
))
ETC_RESPONSE_DEBUG_USER_INFO_COLOR = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_RESPONSE_OUTPUT_USER_INFO_COLOR',
    Fore.MAGENTA if COLORAMA_PRESENT else '',
))

# Enabling/disabling output of specific sections.
ETC_INCLUDE_RESPONSE_DEBUG_URL = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_URL',
    True,
))
ETC_INCLUDE_RESPONSE_DEBUG_CONTENT = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_CONTENT',
    True,
))
ETC_INCLUDE_RESPONSE_DEBUG_HEADER = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_HEADER',
    True,
))
ETC_INCLUDE_RESPONSE_DEBUG_CONTEXT = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_CONTEXT',
    True,
))
ETC_INCLUDE_RESPONSE_DEBUG_SESSION = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_SESSION',
    True,
))
ETC_INCLUDE_RESPONSE_DEBUG_MESSAGES = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_MESSAGES',
    True,
))
ETC_INCLUDE_RESPONSE_DEBUG_FORMS = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_FORMS',
    True,
))
ETC_INCLUDE_RESPONSE_DEBUG_USER_INFO = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_INCLUDE_RESPONSE_DEBUG_USER_INFO',
    True,
))


# Void element list as defined at:
# https://www.w3.org/TR/2011/WD-html-markup-20110113/syntax.html#void-element
# TLDR: A "void element" is an HTML element that does not require a closing tag.
# Used in the "repeating element" logic.
VOID_ELEMENT_LIST = [
    'area',
    'base',
    'br',
    'col',
    'command',
    'embed',
    'hr',
    'img',
    'input',
    'keygen',
    'link',
    'meta',
    'param',
    'source',
    'track',
    'wbr',
]


# Indicates whether the additional debug information should be output.
ETC_DEBUG_PRINT = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_DEBUG_PRINT',
    True,
))
# Indicates whether partial matches are allowed for site title assertions.
# IE: Will pass as long as provided text appears as a subsection of the full site title.
# Defaults to needing exact match.
ETC_ALLOW_TITLE_PARTIALS = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_ALLOW_TITLE_PARTIALS',
    False,
))
# Indicates whether partial matches are allowed for message assertions.
# IE: Will pass as long as provided text appears as a subsection of one or more messages.
# Defaults to needing exact match.
ETC_ALLOW_MESSAGE_PARTIALS = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_ALLOW_MESSAGE_PARTIALS',
    False,
))
# Indicates whether tests fail when there are messages in the response that were not explicitly tested for.
# Aka, when doing ANY message assertion, if this is True, then ALL messages in response need to be checked.
ETC_MATCH_ALL_CONTEXT_MESSAGES = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_MATCH_ALL_CONTEXT_MESSAGES',
    False,
))


# Controls if test-users should be automatically generated or not.
ETC_AUTO_GENERATE_USERS = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_AUTO_GENERATE_USERS',
    True,
))
# Controls what level of strictness UnitTest requests have for users.
ETC_REQUEST_USER_STRICTNESS = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_REQUEST_USER_STRICTNESS',
    'anonymous',
)).strip().lower()
# Allows incorporating package with non-standard user identifiers.
# Such as the common case of using a user email as an identifier, instead of a username.
ETC_USER_MODEL_IDENTIFIER = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_USER_MODEL_IDENTIFIER',
    'username',
))
# The identifier used for the auto-generated "superuser" user.
ETC_DEFAULT_SUPER_USER_IDENTIFIER = getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_DEFAULT_SUPER_USER_IDENTIFIER',
    None,
)
# The identifier used for the auto-generated "admin" user.
ETC_DEFAULT_ADMIN_USER_IDENTIFIER = getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_DEFAULT_ADMIN_USER_IDENTIFIER',
    None,
)
# The identifier used for the auto-generated "standard" user.
ETC_DEFAULT_STANDARD_USER_IDENTIFIER = getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_DEFAULT_STANDARD_USER_IDENTIFIER',
    None,
)
# The identifier used for the auto-generated "inactive" user.
ETC_DEFAULT_INACTIVE_USER_IDENTIFIER = getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_DEFAULT_INACTIVE_USER_IDENTIFIER',
    None,
)
# The default password used for auto-generated users.
ETC_DEFAULT_USER_PASSWORD = str(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_DEFAULT_PASSWORD',
    'password',
))
# Indicates if auto-generated users should get pretend "real" first/last name values.
ETC_GENERATE_USERS_WITH_REAL_NAMES = bool(getattr(
    settings,
    'DJANGO_EXPANDED_TESTCASES_GENERATE_USERS_WITH_REAL_NAMES',
    False,
))


# region User Identifiers

# Validate ETC_REQUEST_USER_STRICTNESS setting.
if ETC_REQUEST_USER_STRICTNESS not in ['anonymous', 'relaxed', 'strict']:
    raise ValueError(
        'Invalid value provided for EXPANDED_TEST_CASES_REQUEST_USER_STRICTNESS setting. '
        'Must be one of: ["anonymous", "relaxed", "strict"].'
    )
# Validate combination of ETC_REQUEST_USER_STRICTNESS and ETC_AUTO_GENERATE_USERS settings.
elif ETC_REQUEST_USER_STRICTNESS == 'relaxed' and not ETC_AUTO_GENERATE_USERS:
    raise ValueError('When ETC_REQUEST_USER_STRICTNESS is set to "relaxed", ETC_AUTO_GENERATE_USERS must be True.')

# Set default identifier value, based on either provided value or common user identifier types.
default_superuser_identifier = None
default_admin_identifier = None
default_inactive_identifier = None
default_user_identifier = None

if ETC_USER_MODEL_IDENTIFIER == 'username':
    # Set default identifiers in username format.
    if ETC_DEFAULT_SUPER_USER_IDENTIFIER is None:
        default_superuser_identifier = 'test_superuser'
    if ETC_DEFAULT_ADMIN_USER_IDENTIFIER is None:
        default_admin_identifier = 'test_admin'
    if ETC_DEFAULT_INACTIVE_USER_IDENTIFIER is None:
        default_inactive_identifier = 'test_inactive'
    if ETC_DEFAULT_STANDARD_USER_IDENTIFIER is None:
        default_user_identifier = 'test_user'

elif ETC_USER_MODEL_IDENTIFIER == 'email':
    # Set default identifiers in email format.
    if ETC_DEFAULT_SUPER_USER_IDENTIFIER is None:
        default_superuser_identifier = 'test_superuser@example.com'
    if ETC_DEFAULT_ADMIN_USER_IDENTIFIER is None:
        default_admin_identifier = 'test_admin@example.com'
    if ETC_DEFAULT_INACTIVE_USER_IDENTIFIER is None:
        default_inactive_identifier = 'test_inactive@example.com'
    if ETC_DEFAULT_STANDARD_USER_IDENTIFIER is None:
        default_user_identifier = 'test_user@example.com'

# Handle any identifiers that have not yet been set by this point.
if default_superuser_identifier is None:
    default_superuser_identifier = str(ETC_DEFAULT_SUPER_USER_IDENTIFIER)
if default_admin_identifier is None:
    default_admin_identifier = str(ETC_DEFAULT_ADMIN_USER_IDENTIFIER)
if default_inactive_identifier is None:
    default_inactive_identifier = str(ETC_DEFAULT_INACTIVE_USER_IDENTIFIER)
if default_user_identifier is None:
    default_user_identifier = str(ETC_DEFAULT_STANDARD_USER_IDENTIFIER)

ETC_DEFAULT_SUPER_USER_IDENTIFIER = default_superuser_identifier
ETC_DEFAULT_ADMIN_USER_IDENTIFIER = default_admin_identifier
ETC_DEFAULT_INACTIVE_USER_IDENTIFIER = default_inactive_identifier
ETC_DEFAULT_STANDARD_USER_IDENTIFIER = default_user_identifier

# endregion User Identifiers



"""
Core testing logic that pertains to handling Response objects.
"""

# System Imports.
import logging, re
from urllib.parse import parse_qs

# Third-Party Imports.
from bs4 import BeautifulSoup
from django.conf import settings
from django.contrib.auth import get_user_model
from django.http.response import HttpResponseBase

# Internal Imports.
from . import CoreTestCaseMixin
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django_expanded_test_cases.constants import (
    ETC_OUTPUT_ERROR_COLOR,
    ETC_RESPONSE_DEBUG_CONTENT_COLOR,
    ETC_RESPONSE_DEBUG_HEADER_COLOR,
    ETC_RESPONSE_DEBUG_CONTEXT_COLOR,
    ETC_RESPONSE_DEBUG_MESSAGE_COLOR,
    ETC_RESPONSE_DEBUG_SESSION_COLOR,
    ETC_RESPONSE_DEBUG_URL_COLOR,
    ETC_RESPONSE_DEBUG_FORM_COLOR,
    ETC_RESPONSE_DEBUG_USER_INFO_COLOR,
    ETC_OUTPUT_EMPHASIS_COLOR,
)


# Initialize logging.
logger = logging.getLogger(__name__)


class ResponseTestCaseMixin(CoreTestCaseMixin):
    """Includes testing logic used in handling Response objects."""

    @classmethod
    def set_up_class(cls, debug_print=None):
        """
        Acts as the equivalent of the UnitTesting "setUpClass()" function.

        However, since this is not inheriting from a given TestCase, calling the literal function
        here would override instead.

        :param debug_print: Optional bool that indicates if debug output should print to console.
                            Param overrides setting value if both param and setting are set.
        """
        # Run parent setup logic.
        super().set_up_class(debug_print=debug_print)

    # region Debug Output Functions

    def show_debug_url(self, url):
        """Prints debug url output."""

        # Ensure url is in consistent format.
        url = self.standardize_url(url, append_root=True)
        message = 'Attempting to access url "{0}"'.format(url)

        self._debug_print('\n\n')
        self._debug_print('{0}'.format('-' * len(message)), fore=ETC_RESPONSE_DEBUG_URL_COLOR, style=ETC_OUTPUT_EMPHASIS_COLOR)
        self._debug_print(message, fore=ETC_RESPONSE_DEBUG_URL_COLOR, style=ETC_OUTPUT_EMPHASIS_COLOR)
        self._debug_print('{0}'.format('-' * len(message)), fore=ETC_RESPONSE_DEBUG_URL_COLOR, style=ETC_OUTPUT_EMPHASIS_COLOR)

    def show_debug_content(self, response_content):
        """Prints debug response page output."""

        # Handle for potential param types.
        if isinstance(response_content, HttpResponseBase):
            response_content = response_content.content.decode('utf-8')
        elif isinstance(response_content, bytes):
            response_content = response_content.decode('utf-8')

        # Standardize output for easier analysis.
        response_content = self.standardize_characters(response_content)
        response_content = self.standardize_newlines(response_content)

        self._debug_print()
        self._debug_print(
            '{0} {1} {0}'.format('=' * 10, 'response.content'),
            fore=ETC_RESPONSE_DEBUG_CONTENT_COLOR,
            style=ETC_OUTPUT_EMPHASIS_COLOR,
        )

        # Print out data, if present.
        if response_content:
            self._debug_print(response_content)
            self._debug_print()

    def show_debug_headers(self, response_headers):
        """Prints debug response header data."""

        # Handle for potential param types.
        if isinstance(response_headers, HttpResponseBase):
            # Actual attr name seemed to change based on settings definitions.
            # Unsure of why though? May want to figure out details at later date.
            # This should account for both instances I'm pretty sure I encountered, though.
            if hasattr(response_headers, 'headers'):
                response_headers = response_headers.headers
            elif hasattr(response_headers, '_headers'):
                response_headers = response_headers._headers
            else:
                response_headers = None

        self._debug_print()
        self._debug_print(
            '{0} {1} {0}'.format('=' * 10, 'response.headers'),
            fore=ETC_RESPONSE_DEBUG_HEADER_COLOR,
            style=ETC_OUTPUT_EMPHASIS_COLOR,
        )

        # Print out data, if present.
        if response_headers is not None and len(response_headers) > 0:
            for key, value in response_headers.items():
                self._debug_print('    * "{0}": "{1}"'.format(key, value), fore=ETC_RESPONSE_DEBUG_HEADER_COLOR)
        else:
            self._debug_print('    No response headers found.')
        self._debug_print()

    def show_debug_context(self, response_context):
        """Prints debug response context data."""

        # Handle for potential param types.
        if isinstance(response_context, HttpResponseBase):
            response_context = response_context.context or {}

        if response_context is not None:
            # Attempt to access key object. If fails, attempt to generate dict of values.
            try:
                response_context.keys()
            except AttributeError:
                # Handling for:
                #     * django.template.context.RequestContext
                #     * django.template.context.Context
                # No guarantee this will work for other aribtrary types.
                # Handle as they come up.
                temp_dict = {}
                for context in response_context:
                    temp_dict = {**temp_dict, **context}
                response_context = temp_dict

        self._debug_print()
        self._debug_print(
            '{0} {1} {0}'.format('=' * 10, 'response.context'),
            fore=ETC_RESPONSE_DEBUG_CONTEXT_COLOR,
            style=ETC_OUTPUT_EMPHASIS_COLOR,
        )

        # NOTE: Response context object is strange, in that it's basically a dictionary,
        # and it allows .keys() but not .values(). Thus, iterate on keys only and pull respective value.
        if response_context is not None and len(response_context.keys()) > 0:
            # pass
            for key in response_context.keys():
                context_value = str(response_context.get(key))
                # Truncate display if very long.
                if len(context_value) > 80:
                    context_value = '"{0}"..."{1}"'.format(context_value[:40], context_value[-40:])
                self._debug_print('    * {0}: {1}'.format(key, context_value), fore=ETC_RESPONSE_DEBUG_CONTEXT_COLOR)
        else:
            self._debug_print('    No context data found.', fore=ETC_RESPONSE_DEBUG_CONTEXT_COLOR)
        self._debug_print()

    def show_debug_session_data(self, client):
        """Prints debug response session data."""

        # Handle for potential param types.
        if isinstance(client, HttpResponseBase):
            client = client.client

        self._debug_print()
        self._debug_print(
            '{0} {1} {0}'.format('=' * 10, 'client.session'),
            fore=ETC_RESPONSE_DEBUG_SESSION_COLOR,
            style=ETC_OUTPUT_EMPHASIS_COLOR,
        )

        if client is not None and len(client.session.items()) > 0:
            for key, value in client.session.items():
                self._debug_print('    * {0}: {1}'.format(key, value), fore=ETC_RESPONSE_DEBUG_SESSION_COLOR)
        else:
            self._debug_print('    No session data found.', fore=ETC_RESPONSE_DEBUG_SESSION_COLOR)
        self._debug_print()

    def show_debug_messages(self, response_context):
        """Prints debug response message data."""

        # Handle for potential param types.
        if isinstance(response_context, HttpResponseBase):
            response_context = response_context.context

        self._debug_print()
        self._debug_print(
            '{0} {1} {0}'.format('=' * 10, 'response.context["messages"]'),
            fore=ETC_RESPONSE_DEBUG_MESSAGE_COLOR,
            style=ETC_OUTPUT_EMPHASIS_COLOR,
        )

        # Print out data, if present.
        if response_context is not None and 'messages' in response_context:
            messages = response_context['messages']
            if len(messages) > 0:
                for message in messages:
                    self._debug_print('    * "{0}"'.format(message), fore=ETC_RESPONSE_DEBUG_MESSAGE_COLOR)
            else:
                self._debug_print('    No context messages found.', fore=ETC_RESPONSE_DEBUG_MESSAGE_COLOR)
        else:
            self._debug_print('    No context messages found.', fore=ETC_RESPONSE_DEBUG_MESSAGE_COLOR)
        self._debug_print()

    def show_debug_form_data(self, response_context):
        """Prints debug response form data."""

        # Handle for potential param types.
        if isinstance(response_context, HttpResponseBase):
            response_context = response_context.context

        self._debug_print()
        self._debug_print(
            '{0} {1} {0}'.format('=' * 10, 'Form Data'),
            fore=ETC_RESPONSE_DEBUG_FORM_COLOR,
            style=ETC_OUTPUT_EMPHASIS_COLOR,
        )

        # Check if form or formset data is actually present.
        if response_context is not None and ('form' in response_context or 'formset' in response_context):
            # Form or formset present on page.

            # Attempt to get form data.
            form_present = False
            if 'form' in response_context:
                form_present = True
                form = response_context['form']

                # Print general form data.
                self._debug_print('    Provided Form Fields:', fore=ETC_RESPONSE_DEBUG_FORM_COLOR)
                fields_submitted = False
                for key, value in form.data.items():
                    self._debug_print('        {0}: {1}'.format(key, value), fore=ETC_RESPONSE_DEBUG_FORM_COLOR)
                    fields_submitted = True
                if not fields_submitted:
                    self._debug_print('        No form field data submitted.', fore=ETC_RESPONSE_DEBUG_FORM_COLOR)

                # Print form data errors if present.
                if not form.is_valid():
                    self._debug_print()
                    if len(form.errors) > 0 or len(form.non_field_errors()) > 0:
                        self._debug_print('    Form Invalid:'.format(not form.is_valid()), fore=ETC_RESPONSE_DEBUG_FORM_COLOR)
                        if len(form.non_field_errors()) > 0:
                            self._debug_print('        Non-field Frrors:', fore=ETC_RESPONSE_DEBUG_FORM_COLOR)
                            for error in form.non_field_errors():
                                self._debug_print('            {0}'.format(error), fore=ETC_RESPONSE_DEBUG_FORM_COLOR)

                        if len(form.errors) > 0:
                            self._debug_print('        Field Errors:', fore=ETC_RESPONSE_DEBUG_FORM_COLOR)
                            for error in form.errors:
                                self._debug_print('            {0}'.format(error), fore=ETC_RESPONSE_DEBUG_FORM_COLOR)

                else:
                    self._debug_print('    Form found and valid.', fore=ETC_RESPONSE_DEBUG_FORM_COLOR)

            # Attempt to get formset data.
            if 'formset' in response_context:
                if form_present:
                    self._debug_print()
                formset = response_context['formset']

                for form in formset:
                    self._debug_print('Form(set) Errors:')
                    for error in form.non_field_errors():
                        self._debug_print('    {0}'.format(error), fore=ETC_RESPONSE_DEBUG_FORM_COLOR)
                    for error in form.errors:
                        self._debug_print('    {0}'.format(error), fore=ETC_RESPONSE_DEBUG_FORM_COLOR)
        else:
            # No identifiable form or formset data present on page.
            self._debug_print('    No form data found.', fore=ETC_RESPONSE_DEBUG_FORM_COLOR)

        self._debug_print()

    def show_debug_user_info(self, user):
        """Prints debug user data."""

        # Imported here to prevent potential "Apps aren't loaded yet" error.
        from django.contrib.auth.models import AnonymousUser

        self._debug_print()
        self._debug_print(
            '{0} {1} {0}'.format('=' * 10, 'User Info'),
            fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR,
            style=ETC_OUTPUT_EMPHASIS_COLOR,
        )

        # Only proceed if we got a proper user model.
        if isinstance(user, get_user_model()):

            # General user information.
            self._debug_print('    * pk: "{0}"'.format(user.pk), fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR)
            self._debug_print('    * Username: "{0}"'.format(user.username), fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR)
            self._debug_print('    * First: "{0}"'.format(user.first_name), fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR)
            self._debug_print('    * Last: "{0}"'.format(user.last_name), fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR)
            self._debug_print('    * Email: "{0}"'.format(user.email), fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR)

            # User auth data.
            self._debug_print(
                '    * is_authenticated: {0}'.format(user.is_authenticated),
                fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR,
            )

            # User groups.
            self._debug_print('    * User Groups: {0}'.format(user.groups.all()), fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR)

            # User permissions.
            self._debug_print(
                '    * User Permissions: {0}'.format(user.user_permissions.all()),
                fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR,
            )

        elif isinstance(user, AnonymousUser):

            self._debug_print('    Anonymous user. No user is logged in.', fore=ETC_RESPONSE_DEBUG_USER_INFO_COLOR)

        else:
            self._debug_print('    * Invalid user "{0}" of type "{1}". Expected "{2}".'.format(
                user,
                type(user),
                type(get_user_model()),
            ), fore=ETC_OUTPUT_ERROR_COLOR)
        self._debug_print()
        self._debug_print()

    # endregion Debug Output Functions.

    def standardize_url(self, url, url_args=None, url_kwargs=None, url_query_params=None, append_root=True):
        """Attempts to standardize URL value, such as in event url is in format for reverse() function.

        :param url: Url value to attempt to parse and standardize.
        :param url_args: Additional "args" to pass for reverse() function, if applicable.
        :param url_kwargs: Additional "kwargs" to pass for reverse() function, if applicable.
        :param append_root: Bool indicating if "site root" should be included in url (if not already).
        :return: Attempt at fully-formatted url.
        """

        # Handle mutable data defaults.
        url_args = url_args or ()
        url_kwargs = url_kwargs or {}
        url_query_params = url_query_params or {}

        # Preprocess all potential url values.
        url = str(url).strip()

        # Attempt to get reverse of provided url.
        try:
            url = reverse(url, args=url_args, kwargs=url_kwargs)

            # Provide warning based on APPEND_SLASH setting.
            # See https://stackoverflow.com/a/42213107 for discussion on why this setting exists
            # as well as how Django generally handles url composition.
            if settings.APPEND_SLASH:
                if len(url) > 0 and url[-1] != '/':
                    warn_msg = (
                        'Django setting APPEND_SLASH is set to True, '
                        'but url did not resolve with trailing slash. '
                        'This may cause UnitTests with ETC to fail. Url was: {0}'
                    ).format(url)
                    logging.warning(warn_msg)
            else:
                if len(url) > 0 and url[-1] == '/':
                    warn_msg = (
                        'Django setting APPEND_SLASH is set to False, '
                        'but url resolved with trailing slash. '
                        'This may cause UnitTests with ETC to fail. Url was: {0}'
                    ).format(url)
                    logging.warning(warn_msg)

        except NoReverseMatch:
            # Could not find as reverse. Assume is literal url.

            # Trim any known extra values on literal url str.
            url = str(url).strip().lstrip('/').rstrip('/')
            if url.startswith(self.site_root_url):
                url = url.lstrip(self.site_root_url)
            elif url.startswith('127.0.0.1'):
                url = url.lstrip('127.0.0.1')

            # Check if empty url.
            if len(url) == 0:
                # Empty url. Populate to single slash.
                url = '/'
            else:
                # Non-empty url.

                # Handle for prepend slash.
                # Only apply if not already present.
                if not url.startswith('/'):
                    url = '/{0}'.format(url)

                # Handle if GET args are included within url.
                split_url = url.split('?')
                if len(split_url) == 1:
                    # No GET args found.
                    url = split_url[0]
                else:
                    # Url GET args found. Temporarily split to prevent errors.
                    url = split_url.pop(0)

                    # Convert query params to expected dictionary format for passing as proper kwargs.
                    url_args = '{0}'.format('?'.join(x for x in split_url))
                    temp_dict = parse_qs(url_args)
                    for key, value in temp_dict.items():
                        url_query_params[key] = value[0]

                # Handle for append slash.
                # Only apply if not already present and settings.APPEND_SLASH is true.
                if settings.APPEND_SLASH and not url.endswith('/'):
                    url = '{0}/'.format(url)

        if url_query_params:
            url = self.generate_get_url(url, **url_query_params)

        # Finally, prepend site root to url, if applicable.
        if append_root:
            url = '{0}{1}'.format(self.site_root_url, url)

        # Return final full url.
        return url

    def standardize_html_tags(self, value):
        """Standardizes spacing around html-esque elements, to remove unnecessary spacing.

        For example, "<h1> MyHeader </h1>" will simplify down to "<h1>MyHeader</h1>".

        Ensures that the actual expected value can be directly checked,
        instead of trying to account for extraneous template whitespacing.

        :param value: Str value to standardize.
        :return: Sanitized str.
        """
        value = str(value)

        # Remove any whitespace around an opening html bracket ( < ).
        value = re.sub(r'(( )*)<(( )*)', '<', value)

        # Remove any whitespace around a closing html bracket ( > ).
        value = re.sub(r'(( )*)>(( )*)', '>', value)

        # Remove any whitespace around an opening array bracket ( [ ).
        value = re.sub(r'(( )*)\[(( )*)', '[', value)

        # Remove any whitespace around a closing array bracket ( ] ).
        value = re.sub(r'(( )*)](( )*)', ']', value)

        # Remove any whitespace around an opening dict bracket ( { ).
        value = re.sub(r'(( )*){(( )*)', '{', value)

        # Remove any whitespace around a closing dict bracket ( } ).
        value = re.sub(r'(( )*)}(( )*)', '}', value)

        return value

    def get_minimized_response_content(self, response, strip_newlines=False):
        """Returns response content, but with minimalistic formatting.

        For example, will trim all newline characters and repeating spaces.
        Also will standardize common characters to output in a single format.

        :param response: Response object or response content to parse.
        :param strip_newlines: Optional bool, indicating if all newlines should be converted into spaces,
            for extra-minimalistic formatting.
            Generally True for UnitTest direct comparison. False for console output/human examination.
        :return: Formatted response content.
        """
        # Handle for provided response types.
        if isinstance(response, HttpResponseBase):
            response_content = response.content.decode('utf-8')
        else:
            response_content = str(response)

        # Standardize basic characters, for easier comparison.
        response_content = self.standardize_characters(response_content)

        # Trim all extra whitespaces, for easier comparison.
        if strip_newlines:
            # All whitespace is converted into a single space.
            # Newlines are also converted into a single space.
            response_content = self.standardize_whitespace(response_content)
        else:
            # All whitespace is converted into a single space.
            # Repeating newlines are condensed down into a single newline.
            response_content = self.standardize_newlines(response_content)

        # Further trim whitespace around specific characters, for easier comparison.
        response_content = self.standardize_html_tags(response_content)

        return response_content

    # region Html Search Functions

    def find_elements_by_tag(self, content, element):
        """Finds all HTML elements that match the provided element tag.

        :param content: Content to search through.
        :param element: Html element to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Sanitize provided element value. We don't care how the user was provided the syntax.
        element = self.get_minimized_response_content(element)
        element = element.lstrip('<').rstrip('>').strip('/').strip()

        # Search for all matching elements.
        soup = BeautifulSoup(content, 'html.parser')
        elements = soup.find_all(name=element)
        element_list = [self.get_minimized_response_content(element.prettify()) for element in elements]

        # Validate parsed value.
        if not len(element_list) > 0:
            self.fail(f'Unable to find element "<{element}>" in content. Provided content was:\n{content}')

        # Return found list.
        return element_list

    def find_element_by_tag(self, content, element):
        """Finds first HTML element that matches the provided element tag.

        :param content: Content to search through.
        :param element: Html element to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Sanitize provided element value. We don't care how the user was provided the syntax.
        element = self.get_minimized_response_content(element)
        element = element.lstrip('<').rstrip('>').strip('/').strip()

        # Call parent function logic.
        element_list = self.find_elements_by_tag(content, element)

        # Verify only one value was found.
        if len(element_list) > 1:
            self.fail(
                f'Found multiple instances of "<{element}>" element. Expected only one instance. '
                f'Content was:\n{content}'
             )

        # Return found item.
        return element_list[0]

    def find_elements_by_id(self, content, element_id):
        """Finds all HTML elements that match the provided id.

        :param content: Content to search through.
        :param element_id: Element id to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Search for all matching elements.
        soup = BeautifulSoup(content, 'html.parser')
        elements = soup.find_all(id=element_id)
        element_list = [self.get_minimized_response_content(element.prettify()) for element in elements]

        # Verify one or more values were found.
        if not len(element_list) > 0:
            self.fail(f'Unable to find id "{element_id}" in content. Provided content was:\n{content}')

        # Provide warning if two or more values were found.
        if len(element_list) > 1:
            logger.warning(
                'It\'s considered bad practice to have multiple matching id tags in one html response. '
                'Consider refactoring elements to keep each id unique.\n'
                'Found {0} total elements with id of "{1}".'.format(len(element_list), element_id)
            )

        # Return found values.
        return element_list

    def find_element_by_id(self, content, element_id):
        """Finds first HTML element that matches the provided id.

        :param content: Content to search through.
        :param element_id: Element id to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Call parent function logic.
        element_list = self.find_elements_by_id(content, element_id)

        # Verify only one value was found.
        if len(element_list) > 1:
            self.fail(
                f'Found multiple instances of "{element_id}" id. Expected only one instance. '
                f'Content was:\n{content}'
            )

        # Return found item.
        return element_list[0]

    def find_elements_by_class(self, content, css_class):
        """Finds all HTML elements that match the provided css class.

        :param content: Content to search through.
        :param css_class: Css class to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Search for all matching elements.
        soup = BeautifulSoup(content, 'html.parser')
        elements = soup.find_all(class_=css_class)
        element_list = [self.get_minimized_response_content(element.prettify()) for element in elements]

        # Verify one or more values were found.
        if not len(element_list) > 0:
            self.fail(f'Unable to find class "{css_class}" in content. Provided content was:\n{content}')

        # Return found values.
        return element_list

    def find_element_by_class(self, content, css_class):
        """Finds first HTML element that matches the provided css class.

        :param content: Content to search through.
        :param css_class: Css class to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Call parent function logic.
        element_list = self.find_elements_by_class(content, css_class)

        # Verify only one value was found.
        if len(element_list) > 1:
            self.fail(
                f'Found multiple instances of "{css_class}" class. Expected only one instance. '
                f'Content was:\n{content}'
            )

        # Return found item.
        return element_list[0]

    def find_elements_by_css_selector(self, content, css_selector):
        """Finds all HTML elements that match the provided css selector.

        :param content: Content to search through.
        :param css_selector: Css selector to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Search for all matching elements.
        soup = BeautifulSoup(content, 'html.parser')
        elements = soup.select(css_selector)
        element_list = [self.get_minimized_response_content(element.prettify()) for element in elements]

        # Verify one or more values were found.
        if not len(element_list) > 0:
            self.fail(f'Unable to find css selector "{css_selector}" in content. Provided content was:\n{content}')

        # Return found values.
        return element_list

    def find_element_by_css_selector(self, content, css_selector):
        """Finds first HTML element that matches the provided css selector.

        :param content: Content to search through.
        :param css_selector: Css selector to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Call parent function logic.
        element_list = self.find_elements_by_css_selector(content, css_selector)

        # Verify only one value was found.
        if len(element_list) > 1:
            self.fail(
                f'Found multiple instances of "{css_selector}" css selector. Expected only one instance. '
                f'Content was:\n{content}'
            )

        # Return found item.
        return element_list[0]

    def find_elements_by_data_attribute(self, content, data_attribute, data_value):
        """Finds all HTML elements that match the provided data attribute and data value.

        :param content: Content to search through.
        :param data_attribute: The key of the data attribute to search for.
        :param data_value: The value of the data attribute to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Search for all matching elements.
        soup = BeautifulSoup(content, 'html.parser')
        attr_dict = {data_attribute: data_value}
        elements = soup.find_all(attrs=attr_dict)
        element_list = [self.get_minimized_response_content(element.prettify()) for element in elements]

        # Verify one or more values were found.
        if not len(element_list) > 0:
            self.fail(
                f'Unable to find data attribute "{data_attribute}" with value "{data_value}" in content. Provided '
                f'content was:\n{content}'
            )

        # Return found values.
        return element_list

    def find_element_by_data_attribute(self, content, data_attribute, data_value):
        """Finds first HTML element that matches the provided data attribute and data value.

        :param content: Content to search through.
        :param data_attribute: The key of the data attribute to search for.
        :param data_value: The value of the data attribute to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Call parent function logic.
        element_list = self.find_elements_by_data_attribute(content, data_attribute, data_value)

        # Verify only one value was found.
        if len(element_list) > 1:
            self.fail(
                f'Found multiple instances of "{data_attribute}" data attribute with value "{data_value}". '
                f'Expected only one instance. Content was:\n{content}'
            )

        # Return found item.
        return element_list[0]

    def find_elements_by_name(self, content, element_name):
        """Finds all HTML elements that match the provided name attribute.

        :param content: Content to search through.
        :param element_name: Element name to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Search for all matching elements.
        soup = BeautifulSoup(content, 'html.parser')
        attr_dict = {'name': element_name}
        elements = soup.find_all(attrs=attr_dict)
        element_list = [self.get_minimized_response_content(element.prettify()) for element in elements]

        # Verify one or more values were found.
        if not len(element_list) > 0:
            self.fail(f'Unable to find name "{element_name}" in content. Provided content was:\n{content}')

        # Return found values.
        return element_list

    def find_element_by_name(self, content, element_name):
        """Finds first HTML element that matches the provided name attribute.

        :param content: Content to search through.
        :param element_name: Element name to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Call parent function logic.
        element_list = self.find_elements_by_name(content, element_name)

        # Verify only one value was found.
        if len(element_list) > 1:
            self.fail(
                f'Found multiple instances of "{element_name}" name. Expected only one instance. '
                f'Content was:\n{content}'
            )

        # Return found item.
        return element_list[0]

    def find_elements_by_link_text(self, content, link_text):
        """Finds all HTML elements that match the provided link text.

        :param content: Content to search through.
        :param link_text: Link text to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Search for all matching elements.
        soup = BeautifulSoup(content, 'html.parser')
        elements = soup.find_all(href=link_text)
        element_list = [self.get_minimized_response_content(element.prettify()) for element in elements]

        # Verify one or more values were found.
        if not len(element_list) > 0:
            self.fail(f'Unable to find link text "{link_text}" in content. Provided content was:\n{content}')

        # Return found values.
        return element_list

    def find_element_by_link_text(self, content, link_text):
        """Finds first HTML element that matches the provided link text.

        :param content: Content to search through.
        :param link_text: Link text to search for.
        """
        # Ensure response content is in expected minimized format.
        content = self.get_minimized_response_content(content)

        # Call parent function logic.
        element_list = self.find_elements_by_link_text(content, link_text)

        # Verify only one value was found.
        if len(element_list) > 1:
            self.fail(
                f'Found multiple instances of "{link_text}" link text. Expected only one instance. '
                f'Content was:\n{content}'
            )

        # Return found item.
        return element_list[0]

    # endregion Html Search Functions


# Define acceptable imports on file.
__all__ = [
    'ResponseTestCaseMixin',
]



"""
Testing logic for views and other multi-part components.
"""

# System Imports.
import re, textwrap

# Third-Party Imports.
from django.conf import settings
from django.core.exceptions import ValidationError
from django.http.response import HttpResponseBase
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch

# Internal Imports.
from .base_test_case import BaseTestCase
from django_expanded_test_cases.constants import (
    ETC_INCLUDE_RESPONSE_DEBUG_URL,
    ETC_INCLUDE_RESPONSE_DEBUG_CONTENT,
    ETC_INCLUDE_RESPONSE_DEBUG_CONTEXT,
    ETC_INCLUDE_RESPONSE_DEBUG_FORMS,
    ETC_INCLUDE_RESPONSE_DEBUG_HEADER,
    ETC_INCLUDE_RESPONSE_DEBUG_MESSAGES,
    ETC_INCLUDE_RESPONSE_DEBUG_SESSION,
    ETC_INCLUDE_RESPONSE_DEBUG_USER_INFO,
    ETC_ALLOW_TITLE_PARTIALS,
    ETC_ALLOW_MESSAGE_PARTIALS,
    ETC_REQUEST_USER_STRICTNESS,
    ETC_DEFAULT_STANDARD_USER_IDENTIFIER,
    ETC_RESPONSE_DEBUG_URL_COLOR,
    ETC_OUTPUT_EMPHASIS_COLOR,
    ETC_OUTPUT_ERROR_COLOR,
    ETC_OUTPUT_RESET_COLOR,
    ETC_OUTPUT_ACTUALS_MATCH_COLOR,
    ETC_OUTPUT_EXPECTED_MATCH_COLOR,
    VOID_ELEMENT_LIST,
)
from django_expanded_test_cases.mixins import ResponseTestCaseMixin


class IntegrationTestCase(BaseTestCase, ResponseTestCaseMixin):
    """Testing functionality for views and other multi-part components."""
    @classmethod
    def setUpClass(cls, *args, debug_print=None, **kwargs):
        # Run parent setup logic.
        super().setUpClass(debug_print=None)

        # Initialize url variables.
        try:
            cls.login_url = reverse(settings.LOGIN_URL)
        except NoReverseMatch:
            # Login url is not defined.
            cls.login_url = None

    # region Custom Assertions

    def assertResponse(
        self,
        url, *args,
        get=True, data=None, secure=True,
        expected_status=200,
        expected_redirect_url=None,
        url_args=None, url_kwargs=None, url_query_params=None,
        redirect_args=None, redirect_kwargs=None, redirect_query_params=None,
        expected_title=None, expected_header=None, expected_messages=None, expected_content=None,
        auto_login=True, user=None, user_permissions=None, user_groups=None, extra_usergen_kwargs=None,
        ignore_content_ordering=False, content_starts_after=None, content_ends_before=None,
        **kwargs,
    ):
        """Verifies the view response object at given URL matches provided parameters.

        At minimum, gets a response object from parsing provided url, then asserts the status code matches.
        Optionally also allows testing:
            * Title - The expected title, aka what displays in the browser tab text at the top of the browser.
            * Header (H1 tag) - The expected H1 header tag on the page.
            * Messages - One or more messages, generated from the Django messages framework.
            * Content - One or more values that should physically appear within html rendering.

        :param url: Url to get response object from.
        :param get: Bool indicating if response is GET or POST. Defaults to GET.
        :param data: Optional dict of items to pass into response generation.
        :param secure: Bool indicating if request should be retrieved as HTTP or HTTPS.
        :param expected_redirect_url: Expected url, after any redirections.
        :param url_args: Values to provide for URL population, in "arg" format.
        :param url_kwargs: Values to provide for URL population, in "kwarg" format.
        :param url_query_params: Query parameter values to provide for URL population.
        :param redirect_args: Values to provide for redirect URL population, in "arg" format.
        :param redirect_kwargs: Values to provide for redirect URL population, in "kwarg" format.
        :param redirect_query_params: Query parameter values to provide for redirect URL population.
        :param expected_status: Expected status code, after any redirections. Default code of 200.
        :param expected_title: Expected page title to verify. Skips title test if left as None.
        :param expected_header: Expected page h1 to verify. Skips header test if left as None.
        :param expected_messages: Expected context messages to verify. Skips message test if left as None.
        :param expected_content: Expected page content elements to verify. Skips content test if left as None.
        :param auto_login: Bool indicating if user should be auto-logged-in.
        :param user: User to log in with, if auto_login is True. Defaults to `test_user`.
        :param user_permissions: Optional permissions to provide to login user.
        :param user_groups: Optional groups to provide to login user.
        :param extra_usergen_kwargs: Optional dictionary of values to pass to _get_login_user__extra_user_auth_setup().
        :param ignore_content_ordering: Bool indicating if ordering should be verified. Defaults to checking ordering.
        :param content_starts_after: The HTML that expected_content should occur after. This HTML and everything
                                     preceding is stripped out of the "search space" for the expected_content value.
        :param content_ends_before: The HTML that expected_content should occur before. This HTML and everything
                                    following is stripped out of the "search space" for the expected_content value.
        """

        # Handle mutable data defaults.
        data = data or {}
        url_args = (*args, *kwargs.pop('args', []), *(url_args or []))
        url_query_params = url_query_params or {}
        redirect_args = (*args, *(redirect_args or []))
        redirect_kwargs = {**kwargs, **(redirect_kwargs or {})}
        redirect_query_params = redirect_query_params or {}
        url_kwargs = {**kwargs.pop('kwargs', {}), **(url_kwargs or {}), **kwargs}
        extra_usergen_kwargs = extra_usergen_kwargs or {}

        # Reset client "user login" state for new response generation.
        self.client.logout()

        # Handle getting user.
        user = self._get_default_request_user(user, auto_login)

        # Run logic to get corresponding response object.
        response = self._get_page_response(
            url,
            get=get,
            data=data,
            secure=secure,
            url_args=url_args,
            url_kwargs=url_kwargs,
            query_params=url_query_params,
            auto_login=auto_login,
            user=user,
            user_permissions=user_permissions,
            user_groups=user_groups,
            extra_usergen_kwargs=extra_usergen_kwargs,
        )

        # Optionally output all debug info for found response.
        if self._debug_print_bool:
            if ETC_INCLUDE_RESPONSE_DEBUG_CONTENT:
                self.show_debug_content(response)
            if ETC_INCLUDE_RESPONSE_DEBUG_HEADER:
                self.show_debug_headers(response)
            if ETC_INCLUDE_RESPONSE_DEBUG_CONTEXT:
                self.show_debug_context(response)
            if ETC_INCLUDE_RESPONSE_DEBUG_SESSION:
                self.show_debug_session_data(response)
            if ETC_INCLUDE_RESPONSE_DEBUG_MESSAGES:
                self.show_debug_messages(response)
            if ETC_INCLUDE_RESPONSE_DEBUG_FORMS:
                self.show_debug_form_data(response)
            if ETC_INCLUDE_RESPONSE_DEBUG_USER_INFO:
                self.show_debug_user_info(response.user)

        # Optional hook for running custom pre-builtin-test logic.
        self._assertResponse__pre_builtin_tests(
            response.url, *args,
            response=response,
            get=get, data=data, secure=secure,
            expected_status=expected_status,
            expected_redirect_url=expected_redirect_url,
            url_args=url_args, url_kwargs=url_kwargs, url_query_params=url_query_params,
            redirect_args=redirect_args, redirect_kwargs=redirect_kwargs, redirect_query_params=redirect_query_params,
            expected_title=expected_title, expected_header=expected_header, expected_messages=expected_messages,
            expected_content=expected_content,
            auto_login=auto_login, user=user, user_permissions=user_permissions, user_groups=user_groups,
            extra_usergen_kwargs=extra_usergen_kwargs,
            ignore_content_ordering=ignore_content_ordering, content_starts_after=content_starts_after,
            content_ends_before=content_ends_before,
            **kwargs,
        )

        # Verify page status code.
        self.assertStatusCode(response, expected_status)

        # Verify page redirect.
        if expected_redirect_url is not None:
            self.assertRedirects(
                response,
                expected_redirect_url,
                redirect_args=redirect_args,
                redirect_kwargs=redirect_kwargs,
                redirect_query_params=redirect_query_params,
            )

        # Verify page title.
        if expected_title is not None:
            self.assertPageTitle(response, expected_title)

        # Verify page header.
        if expected_header is not None:
            self.assertPageHeader(response, expected_header)

        # Verify page messages.
        if expected_messages is not None:
            self.assertContextMessages(response, expected_messages, debug_output=False)

        # Verify page content.
        if expected_content is not None:
            self.assertPageContent(
                response,
                expected_content,
                ignore_ordering=ignore_content_ordering,
                content_starts_after=content_starts_after,
                content_ends_before=content_ends_before,
                debug_output=False,
            )

        # Optional hook for running custom post-builtin-test logic.
        self._assertResponse__post_builtin_tests(
            response.url, *args,
            response=response,
            get=get, data=data, secure=secure,
            expected_status=expected_status,
            expected_redirect_url=expected_redirect_url,
            url_args=url_args, url_kwargs=url_kwargs, url_query_params=url_query_params,
            redirect_args=redirect_args, redirect_kwargs=redirect_kwargs, redirect_query_params=redirect_query_params,
            expected_title=expected_title, expected_header=expected_header, expected_messages=expected_messages,
            expected_content=expected_content,
            auto_login=auto_login, user=user, user_permissions=user_permissions, user_groups=user_groups,
            extra_usergen_kwargs=extra_usergen_kwargs,
            ignore_content_ordering=ignore_content_ordering, content_starts_after=content_starts_after,
            content_ends_before=content_ends_before,
            **kwargs,
        )

        # All assertions passed so far. Return response in case user wants to do further checks.
        return response

    def assertGetResponse(
        self,
        url, *args,
        data=None, secure=True,
        expected_status=200,
        expected_redirect_url=None,
        url_args=None, url_kwargs=None, url_query_params=None,
        redirect_args=None, redirect_kwargs=None, redirect_query_params=None,
        expected_title=None, expected_header=None, expected_messages=None, expected_content=None,
        auto_login=True, user=None, user_permissions=None, user_groups=None, extra_usergen_kwargs=None,
        ignore_content_ordering=False, content_starts_after=None, content_ends_before=None,
        **kwargs,
    ):
        """Verifies a GET response was found at given URL, and matches provided parameters."""

        # Handle mutable data defaults.
        data = data or {}
        url_args = url_args or []
        url_kwargs = url_kwargs or {}
        url_query_params = url_query_params or []
        redirect_args = redirect_args or []
        redirect_kwargs = redirect_kwargs or {}
        redirect_query_params = redirect_query_params or {}
        extra_usergen_kwargs = extra_usergen_kwargs or {}

        # Call base function to handle actual logic.
        return self.assertResponse(
            url,
            *args,
            get=True,
            data=data,
            secure=secure,
            expected_status=expected_status,
            expected_redirect_url=expected_redirect_url,
            url_args=url_args,
            url_kwargs=url_kwargs,
            url_query_params=url_query_params,
            redirect_args=redirect_args,
            redirect_kwargs=redirect_kwargs,
            redirect_query_params=redirect_query_params,
            expected_title=expected_title,
            expected_header=expected_header,
            expected_messages=expected_messages,
            expected_content=expected_content,
            auto_login=auto_login,
            user=user,
            user_permissions=user_permissions,
            user_groups=user_groups,
            extra_usergen_kwargs=extra_usergen_kwargs,
            ignore_content_ordering=ignore_content_ordering,
            content_starts_after=content_starts_after,
            content_ends_before=content_ends_before,
            **kwargs,
        )

    def assertPostResponse(
        self,
        url, *args,
        data=None, secure=True,
        expected_status=200,
        expected_redirect_url=None,
        url_args=None, url_kwargs=None, url_query_params=None,
        redirect_args=None, redirect_kwargs=None, redirect_query_params=None,
        expected_title=None, expected_header=None, expected_messages=None, expected_content=None,
        auto_login=True, user=None, user_permissions=None, user_groups=None, extra_usergen_kwargs=None,
        ignore_content_ordering=False, content_starts_after=None, content_ends_before=None,
        **kwargs,
    ):
        """Verifies a GET response was found at given URL, and matches provided parameters."""

        # Handle mutable data defaults.
        data = data or {}
        url_args = url_args or []
        url_kwargs = url_kwargs or {}
        url_query_params = url_query_params or {}
        redirect_args = redirect_args or []
        redirect_kwargs = redirect_kwargs or {}
        redirect_query_params = redirect_query_params or {}
        extra_usergen_kwargs = extra_usergen_kwargs or {}

        # Forcibly add values to "data" dict, so that POST doesn't validate to empty in view.
        # This guarantees that view serves as POST, like this specific assertion expects.
        if data == {}:
            # Has no values. Forcibly add a single key-value pair.
            data = {'UnitTest': True}

        # Call base function to handle actual logic.
        return self.assertResponse(
            url,
            *args,
            get=False,
            data=data,
            secure=secure,
            expected_status=expected_status,
            expected_redirect_url=expected_redirect_url,
            url_args=url_args,
            url_kwargs=url_kwargs,
            url_query_params=url_query_params,
            redirect_args=redirect_args,
            redirect_kwargs=redirect_kwargs,
            redirect_query_params=redirect_query_params,
            expected_title=expected_title,
            expected_header=expected_header,
            expected_messages=expected_messages,
            expected_content=expected_content,
            auto_login=auto_login,
            user=user,
            user_permissions=user_permissions,
            user_groups=user_groups,
            extra_usergen_kwargs=extra_usergen_kwargs,
            ignore_content_ordering=ignore_content_ordering,
            content_starts_after=content_starts_after,
            content_ends_before=content_ends_before,
            **kwargs,
        )

    def assertRedirects(
        self,
        response, expected_redirect_url, *args,
        redirect_args=None, redirect_kwargs=None, redirect_query_params=None, **kwargs,
    ):
        """Assert that a response redirected to a specific URL and that the redirect URL can be loaded.

        Most functionality is in the default Django assertRedirects() function.
        However, this acts as a wrapper to also:
            * Check that provided response param is a valid Response object. Attempts to generate one if not.
            * Attempt url as reverse, before trying assertion.
        """
        # Handle mutable data defaults.
        redirect_args = redirect_args or []
        redirect_kwargs = redirect_kwargs or {}
        redirect_query_params = redirect_query_params or {}

        # Ensure provided response is actual response.
        if isinstance(response, HttpResponseBase):
            # Is literal response.
            pass
        else:
            # Is not response. Attempt to get.
            response = self._get_page_response(response)

        # Sanitize redirect url.
        expected_redirect_url = self.standardize_url(
            expected_redirect_url,
            url_args=redirect_args,
            url_kwargs=redirect_kwargs,
            url_query_params=redirect_query_params,
            append_root=False,
        )

        # Run assertion on provided value.
        try:
            return super().assertRedirects(response, expected_redirect_url, *args, **kwargs)
        except AssertionError:
            self.fail('Response didn\'t redirect as expected. Response code was {0} (expected 302).'.format(
                response.status_code
            ))

    def assertStatusCode(self, response, expected_status):
        """Verifies the page status code value.

        :param response: Response object to check against.
        :param expected_status: Expected status code, after any redirections.
        """
        # Handle for provided response types.
        if isinstance(response, HttpResponseBase):
            actual_status = response.status_code
        else:
            actual_status = int(response)

        # Check status.
        self.assertEqual(
            actual_status,
            expected_status,
            'Expected status code (after potential redirects) of "{0}". Actual code was "{1}".'.format(
                expected_status,
                actual_status,
            ),
        )

        # Return status in case user wants to run additional logic on it.
        return actual_status

    def assertPageTitle(self, response, expected_title, allow_partials=None):
        """Verifies the page title HTML element.

        Note: Some sites have titles with nested elements.
            Ex: "<title><page> | <app> | <website></title>"

        Thus the "exact_match" bool exists, to allow only testing for a specific unit in the title, instead of
        always having to type the full title for every test.

        :param response: Response object to check against.
        :param expected_title: Expected full string in title HTML element.
        :param allow_partials: Bool indicating if title should be exact match, or partial.
        :return: Parsed out title string.
        """
        # Parse out settings values.
        if allow_partials is None:
            allow_partials = ETC_ALLOW_TITLE_PARTIALS
        else:
            allow_partials = bool(allow_partials)

        # Parse out title element from response.
        actual_title = self.get_page_title(response)

        # Remove title tag from expected value, if present.
        expected_title = str(expected_title).strip()
        if expected_title.startswith('<title>'):
            expected_title = expected_title[7:]
        if expected_title.endswith('</title>'):
            expected_title = expected_title[:-8]
        expected_title = expected_title.strip()

        # Check element.
        err_msg = 'Expected title HTML contents of "{0}" ({1}). Actual value was "{2}".'
        if not allow_partials:
            # Check using exact match.
            if expected_title != actual_title:
                self.fail(err_msg.format(
                    expected_title,
                    'using exact matching',
                    actual_title,
                ))
        else:
            # Check using partial match.
            if expected_title not in actual_title:
                self.fail(err_msg.format(
                    expected_title,
                    'using partial matching',
                    actual_title,
                ))

        # Return title in case user wants to run additional logic on it.
        return actual_title

    def assertPageHeader(self, response, expected_header):
        """Verifies the page H1 header HTML element.

        :param response: Response object to check against.
        :param expected_header: Expected full string in H1 header HTML element.
        :return: Parsed out header string.
        """
        # Parse out H1 header element from response.
        actual_header = self.get_page_header(response)

        # Remove H1 tag from expected value, if present.
        expected_header = str(expected_header).strip()
        if expected_header.startswith('<h1>'):
            expected_header = expected_header[4:]
        if expected_header.endswith('</h1>'):
            expected_header = expected_header[:-5]
        expected_header = expected_header.strip()

        # Check element.
        if expected_header != actual_header:
            self.fail('Expected H1 header HTML contents of "{0}". Actual value was "{1}".'.format(
                expected_header,
                actual_header,
            ))

        # Return header in case user wants to run additional logic on it.
        return actual_header

    def assertContextMessages(self, response, expected_messages, allow_partials=None, debug_output=True):
        """Verifies the context messages.

        :param response: Response object to check against.
        :param expected_messages: Expected string for message data.
        :param allow_partials: Bool indicating if messages should fully match or allow partial matches.
        :param debug_output: Bool indicating if debug output should be shown or not. Used for debugging test failures.
        :return: Parsed out header string.
        """
        if debug_output:
            # Print out actual messages, for debug output.
            self.show_debug_messages(response)

        # Parse out settings values.
        if allow_partials is None:
            allow_partials = ETC_ALLOW_MESSAGE_PARTIALS
        else:
            allow_partials = bool(allow_partials)

        # Parse out message data from response.
        actual_messages = self.get_context_messages(response)

        # Check format of expected value.
        if isinstance(expected_messages, list) or isinstance(expected_messages, tuple):
            # Array of messages passed. Verify each inner value is a str.
            temp_messages = []
            for message in expected_messages:
                message = str(message).strip()
                if len(message) > 0:
                    temp_messages.append(message)
            expected_messages = temp_messages

        elif expected_messages is None:
            # Handle for none type. Not sure why anyone would pass this into the test though.
            expected_messages = []

        elif isinstance(expected_messages, str):
            # For everything else, assume is intended to be a single message.
            message = str(expected_messages).strip()
            expected_messages = []
            if len(message) > 0:
                expected_messages.append(message)

        if len(expected_messages) > 0:
            # For now, we only care about values passed for expected_messages.
            # We ignore any cases where a message exists in the context, but is not explicitly
            # checked by the user in the expected_messages param.

            # One or more messages are expected. Verify they are found.
            for expected_message in expected_messages:
                message_found = False

                # Handle based on partials allowed or not.
                if allow_partials:
                    # Partial message matching is allowed.
                    # Expected value must match or be a substring of a message in context.
                    index = 0
                    while message_found is False and index < len(actual_messages):
                        if expected_message in actual_messages[index]:
                            message_found = True
                        index += 1

                else:
                    # Partial message matching is NOT allowed.
                    # Expected value must exactly match message in context.

                    # Loop through all context messages until found, or all are checked.
                    if expected_message in actual_messages:
                        message_found = True

                # Raise assertion error if not found.
                if not message_found:
                    self.fail('Failed to find message "{0}" in context (using {1} matching).'.format(
                        expected_message,
                        'partial' if allow_partials else 'exact'
                    ))

    def assertPageContent(
        self,
        response, expected_content,
        ignore_ordering=False, content_starts_after=None, content_ends_before=None, debug_output=True,
    ):
        """Verifies the page content html, similar to the built-in assertContains() function.
        The main difference is that Django templating may create large amounts of whitespace in response html,
        often in places where we wouldn't intuitively expect it, when running tests.
        Technically, the built-in assertHTMLEqual() and assertInHTML() functions exist, and probably could accomplish
        the same assertions. But we still need to parse and format full response object, to display for test failure
        debugging. So I'm not sure if it's helpful at that point to use those or use separate assertions like here.
        Perhaps examine more closely at a later date.
        :param response: Response object to check against.
        :param expected_content: Expected full string (or set of strings) of HTML content.
        :param ignore_ordering: Bool indicating if ordering should be verified. Defaults to checking ordering.
        :param content_starts_after: The HTML that expected_content should occur after. This HTML and everything
                                     preceding is stripped out of the "search space" for the expected_content value.
        :param content_ends_before: The HTML that expected_content should occur before. This HTML and everything
                                    following is stripped out of the "search space" for the expected_content value.
        :param debug_output: Bool indicating if debug output should be shown or not. Used for debugging test failures.
        :return: Parsed out and formatted content string.
        """
        if debug_output:
            # Print out actual response content, for debug output.
            self.show_debug_content(response)

        main_err_msg = (
            'Could not find expected content value in response. Provided value was:\n'
            '{0}\n'
        )
        checked_content_str_addon = (
            '\n'
            '\n'
            'Surrounding Checks:\n'
            '{0}'
            '{1}'
            '{2}'
            '{3}'
            '{4}'
        )
        ordering_err_msg = 'Expected content value was found, but ordering of values do not match. Problem value:\n{0}'
        casing_err_msg = (
            'Expected content value was found, but letter capitalization did not match. Expected was:\n'
            '{0}\n'
            '\n'
            'Found was:\n'
            '... {1} ...'
        )

        # Extra setup logic, to sanitize and handle if content_starts_after/content_ends_before variables are defined.
        content_dict = self._trim_response_content(
            response,
            content_starts_after=content_starts_after,
            content_ends_before=content_ends_before,
        )
        sanitized_original_content = content_dict['minimized_content']
        trimmed_original_content = content_dict['truncated_content']

        # Handle possible types.
        if expected_content is None:
            expected_content = ''
        if isinstance(expected_content, list) or isinstance(expected_content, tuple):
            # The expected_content param is an array of items. Verify they all exist on page.
            trimmed_content = trimmed_original_content
            for index in range(len(expected_content)):
                expected = expected_content[index]

                # Update str in event of error.
                updated_checked_content_str_addon = checked_content_str_addon.format(
                    '{1}    * {0}{2}\n'.format(
                        expected_content[index - 2],
                        ETC_OUTPUT_EXPECTED_MATCH_COLOR,
                        ETC_OUTPUT_RESET_COLOR,
                    ) if index >= 2 else '',
                    '{1}    * {0}{2}\n'.format(
                        expected_content[index - 1],
                        ETC_OUTPUT_EXPECTED_MATCH_COLOR,
                        ETC_OUTPUT_RESET_COLOR,
                    ) if index >= 1 else '',
                    '{1}  > * {0}{2}\n'.format(
                        expected_content[index],
                        ETC_OUTPUT_ERROR_COLOR,
                        ETC_OUTPUT_RESET_COLOR,
                    ),
                    '{1}    * {0}{2}\n'.format(
                        expected_content[index + 1],
                        ETC_OUTPUT_ACTUALS_MATCH_COLOR,
                        ETC_OUTPUT_RESET_COLOR,
                    ) if (len(expected_content) - index) > 1 else '',
                    '{1}    * {0}{2}\n'.format(
                        expected_content[index + 2],
                        ETC_OUTPUT_ACTUALS_MATCH_COLOR,
                        ETC_OUTPUT_RESET_COLOR,
                    ) if (len(expected_content) - index) > 2 else '',
                )

                stripped_expected = self.get_minimized_response_content(expected, strip_newlines=True)
                if ignore_ordering:
                    # Ignoring ordering. Check as-is.
                    if stripped_expected not in trimmed_original_content:
                        # Expected value not found in provided content section.
                        display_expected = self.get_minimized_response_content(expected, strip_newlines=False)

                        # Check if due to casing mismatch.
                        if stripped_expected.casefold() in trimmed_original_content.casefold():
                            # Match found when ignoring casing.

                            # Get regex match of actual value, plus 20 characters on each side.
                            search_val = r'((?:[\S\s]{0,20})' + re.escape(stripped_expected) + r'(?:[\S\s]{0,20}))'
                            trimmed_actual = re.search(
                                search_val,
                                trimmed_original_content,
                                flags=re.IGNORECASE,
                            )

                            # Display corresponding error message.
                            self.fail(casing_err_msg.format(display_expected, trimmed_actual.group(0)))

                        else:
                            # Value doesn't exist even after ignoring casing.
                            # Raise message based on content_starts_after/content_ends_before variables.
                            self._assertPageContent(
                                sanitized_original_content,
                                stripped_expected,
                                display_expected,
                                content_starts_after,
                                content_ends_before,
                                main_err_msg,
                                updated_checked_content_str_addon,
                            )
                else:
                    # Verifying ordering.
                    # Attempt initial assertion in provided subsection.
                    if stripped_expected not in trimmed_content:
                        # Failed to find content in subsection. Check full content set.
                        if stripped_expected not in trimmed_original_content:
                            # Expected value not found in provided content section.
                            display_expected = self.get_minimized_response_content(expected, strip_newlines=False)

                            # Check if due to casing mismatch.
                            if stripped_expected.casefold() in trimmed_original_content.casefold():
                                # Match found when ignoring casing.

                                # Get regex match of actual value, plus 20 characters on each side.
                                search_val = r'((?:[\S\s]{0,20})' + re.escape(stripped_expected) + r'(?:[\S\s]{0,20}))'
                                trimmed_actual = re.search(
                                    search_val,
                                    trimmed_original_content,
                                    flags=re.IGNORECASE,
                                )

                                # Display corresponding error message.
                                self.fail(casing_err_msg.format(display_expected, trimmed_actual.group(0)))

                            else:
                                # Value doesn't exist even after ignoring casing.
                                # Raise message based on content_starts_after/content_ends_before variables.
                                self._assertPageContent(
                                    sanitized_original_content,
                                    stripped_expected,
                                    display_expected,
                                    content_starts_after,
                                    content_ends_before,
                                    main_err_msg,
                                    updated_checked_content_str_addon,
                                )

                        # If we made it this far, then item was found in full content, but came after a previous
                        # expected value. Raise error.
                        if updated_checked_content_str_addon:
                            ordering_err_msg += updated_checked_content_str_addon
                        self.fail(ordering_err_msg.format(expected))

                # If we made it this far, then value was found. Handle for ordering.
                if not ignore_ordering:
                    # Ordering is being checked. Strip off first section of matching.
                    trimmed_content = stripped_expected.join(
                        trimmed_content.split(stripped_expected)[1:],
                    )

        else:
            # Not an array of items. Assume is a single str value.
            stripped_expected = self.get_minimized_response_content(expected_content, strip_newlines=True)
            if stripped_expected not in trimmed_original_content:
                # Expected value not found in provided content section.
                display_expected = self.get_minimized_response_content(expected_content, strip_newlines=False)

                # Check if due to casing mismatch.
                if stripped_expected.casefold() in trimmed_original_content.casefold():
                    # Match found when ignoring casing.

                    # Get regex match of actual value, plus 20 characters on each side.
                    search_val = r'((?:[\S\s]{0,20})' + re.escape(stripped_expected) + r'(?:[\S\s]{0,20}))'
                    trimmed_actual = re.search(
                        search_val,
                        trimmed_original_content,
                        flags=re.IGNORECASE,
                    )

                    # Display corresponding error message.
                    self.fail(casing_err_msg.format(display_expected, trimmed_actual.group(0)))

                else:
                    # Value doesn't exist even after ignoring casing.
                    # Raise message based on content_starts_after/content_ends_before variables.
                    self._assertPageContent(
                        sanitized_original_content,
                        stripped_expected,
                        display_expected,
                        content_starts_after,
                        content_ends_before,
                        main_err_msg,
                        '',
                    )

        # Return page content in case user wants to run additional logic on it.
        return trimmed_original_content

    def _assertPageContent(
        self,
        actual_content, minimized_expected, display_expected,
        strip_actual_start, strip_actual_end,
        err_msg, checked_content_str_addon
    ):
        """Internal sub-assertion for assertPageContent() function."""
        strip_err_msg = 'Expected content value was found, but occurred in "{0}" section. Expected was:\n{1}'

        # Check if error was due to content_starts_after/content_ends_before variables.
        found_expected = False
        if strip_actual_start:
            stripped_start_section = str(actual_content.split(strip_actual_start)[0] + strip_actual_start)
            if minimized_expected in stripped_start_section:
                found_expected = 'content_starts_after'

        if strip_actual_end:
            stripped_end_section = str(strip_actual_end + actual_content.split(strip_actual_end)[-1])
            if minimized_expected in stripped_end_section:
                found_expected = 'content_ends_before'

        # Output message based on above searches.
        if found_expected:
            # Content value was in stripped section. Raise corresponding strip message.
            if checked_content_str_addon:
                strip_err_msg += checked_content_str_addon
            self.fail(strip_err_msg.format(found_expected, display_expected))

        else:
            # Content value was physically not present at all. Raise "main" message.
            if checked_content_str_addon:
                err_msg += checked_content_str_addon
            self.fail(err_msg.format(display_expected))

    def assertRepeatingElement(
        self,
        response, expected_repeating_element, repeat_count,
        content_starts_after=None, content_ends_before=None, debug_output=True,
    ):
        """Verifies that a given HTMl element repeats, within a given section of content.

        Note: This expects a full HTML element, including both opening and closing tags.

        :param response: Response object to check against.
        :param expected_repeating_element: The expected repeating HTML element. Ex: <li>, <p>, etc.
        :param repeat_count: Integer indicating how many times the HTML element should repeat.
        :param content_starts_after: The HTML that the element should occur after. This HTML and everything
                                     preceding is stripped out of the "search space" for the expected_content value.
        :param content_ends_before: The HTML that the element should occur before. This HTML and everything
                                    following is stripped out of the "search space" for the expected_content value.
        :param debug_output: Bool indicating if debug output should be shown or not. Used for debugging test failures.
        :return: Parsed out and formatted content string.
        """
        # Standardize provided repeating value.
        expected_repeating_element = self.standardize_characters(expected_repeating_element)

        # Sanitize initial content element.
        repeat_count = int(repeat_count)
        if repeat_count < 1:
            raise ValueError('The assertRepeatingElement() function requires an element occurs one or more times.')
        expected_repeating_element = str(expected_repeating_element).strip().lstrip('<').rstrip('>').rstrip('/').strip()
        expected_repeating_element = expected_repeating_element.lower()
        is_void_element = False
        if expected_repeating_element in VOID_ELEMENT_LIST:
            is_void_element = True

        # Generate expected content set.
        # This is what we pass from this wrapper function to the actual assertion function.
        expected_content = []
        for index in range(repeat_count):
            expected_content.append('<{0}>'.format(expected_repeating_element))

            # Add closing tag if not a void element.
            if not is_void_element:
                expected_content.append('</{0}>'.format(expected_repeating_element))

        # Pass our sanitized values into assertPageContent().
        content_dict = self._trim_response_content(
            response,
            content_starts_after=content_starts_after,
            content_ends_before=content_ends_before,
        )
        truncated_content = content_dict['truncated_content']

        # Check element counts within desired section.
        open_tag_err_msg = 'Expected {0} element{1} tags. Found {2}.'
        close_tag_err_msg = 'Expected {0} element closing tags. Found {1}.'
        open_tag_count = truncated_content.count('<{0}>'.format(expected_repeating_element))

        # Check count of element opening tags.
        try:
            self.assertEqual(open_tag_count, repeat_count)
        except AssertionError:
            self.fail(open_tag_err_msg.format(repeat_count, '' if is_void_element else ' opening', open_tag_count))

        # Check count of element closing tags.
        if not is_void_element:
            close_tag_count = truncated_content.count('</{0}>'.format(expected_repeating_element))
            try:
                self.assertEqual(close_tag_count, repeat_count)
            except AssertionError:
                self.fail(close_tag_err_msg.format(repeat_count, close_tag_count))

        # Run full assertPageContent() to make sure we're thorough (unsure of if this part is needed?).
        return self.assertPageContent(truncated_content, expected_content, debug_output=debug_output)

    # endregion Custom Assertions

    # region Helper Functions

    def _get_default_request_user(self, user, auto_login):
        """
        # We first prioritize a direct value passed into this function (if not the default).
        # Then we prioritize a user saved to the class "self.user" variable (if not the default).
        :param user:
        :param auto_login:
        :return:
        """
        # Django imports here to avoid situational "Apps aren't loaded yet" error.
        from django.contrib.auth.models import AnonymousUser

        # If user is None, attempt to fallback to class-user.
        if user is None:
            class_user = getattr(self, 'user', None)
            if class_user != AnonymousUser and class_user != AnonymousUser():
                user = class_user

        # Proceed if user still None.
        if user is None:
            # Handle if default mode is "anonymous" and no user provided.
            if ETC_REQUEST_USER_STRICTNESS == 'anonymous':
                user = AnonymousUser()

            # Handle if default mode is "relaxed" and no user provided.
            elif ETC_REQUEST_USER_STRICTNESS == 'relaxed':
                user = ETC_DEFAULT_STANDARD_USER_IDENTIFIER

            # Handle if default mode is "strict" and no user provided.
            elif ETC_REQUEST_USER_STRICTNESS == 'strict':
                # Handle for logging in a user.
                if auto_login:
                    raise ValidationError(
                        'ETC_REQUEST_USER_STRICTNESS is set to "strict" but auto_login is True and no user was provided. '
                        'Please either set auto_login to False OR explicitly provide a user to authenticate with.'
                    )

            else:
                raise ValueError(
                    'Invalid value provided for EXPANDED_TEST_CASES_REQUEST_USER_STRICTNESS setting. '
                    'Must be one of: ["anonymous", "relaxed", "strict"].'
                )

        # Use anonymous user if not set to auto-login.
        if not auto_login:
            user = AnonymousUser()

        # Return calculated user.
        return user

    def _get_page_response(
        self,
        url,
        *args,
        get=True, data=None, secure=True,
        url_args=None, url_kwargs=None, query_params=None,
        auto_login=True, user=None, user_permissions=None, user_groups=None, extra_usergen_kwargs=None,
        **kwargs,
    ):
        """Helper function for assertResponse().

        Fully parses provided user url, and returns corresponding response object.

        :param url: Url to get response object from.
        :param get: Bool indicating if response is GET or POST. Defaults to GET.
        :param data: Optional dict of items to pass into response generation.
        :param secure: Bool indicating if request should be retrieved as HTTP or HTTPS.
        :param url_args: Values to provide for URL population, in "arg" format.
        :param url_kwargs: Values to provide for URL population, in "kwarg" format.
        :param auto_login: Bool indicating if User should be "logged in" to client or not.
        :param user_permissions: Set of Django Permissions to give to test user before accessing page.
        :param user_groups: Set of Django PermissionGroups to give to test user before accessing page.
        :param extra_usergen_kwargs: Optional dictionary of values to pass to _get_login_user__extra_user_auth_setup().
        :return: Django response object for provided url.
        """
        # Handle mutable data defaults.
        data = data or {}
        url_args = (*args, *kwargs.pop('args', []), *(url_args or []))
        url_kwargs = {**kwargs.pop('kwargs', {}), **(url_kwargs or {}), **kwargs}
        query_params = query_params or {}
        extra_usergen_kwargs = {**kwargs, **(extra_usergen_kwargs or {})}

        # Validate data types.
        if not isinstance(data, dict):
            raise TypeError('Provided "data" arg must be a dict, for passing into POST requests.')

        # Handle getting user.
        user = self._get_default_request_user(user, auto_login)
        user = self._get_login_user(
            user,
            *args,
            auto_login=auto_login,
            user_permissions=user_permissions,
            user_groups=user_groups,
            **extra_usergen_kwargs,
        )

        # Handle url sanitization.
        url = self.standardize_url(
            url,
            url_args=url_args,
            url_kwargs=url_kwargs,
            url_query_params=query_params,
            append_root=False,
        )
        full_url = '{0}{1}'.format(self.site_root_url, url)
        if ETC_INCLUDE_RESPONSE_DEBUG_URL:
            self.show_debug_url(full_url)

        # Get response object.
        if bool(get):
            response = self.client.get(url, data=data, secure=secure, follow=True)
        else:
            response = self.client.post(url, data=data, secure=secure, follow=True)

        # Update response object with additional useful values for further testing/analysis.
        response.url = url
        response.full_url = full_url
        response.user = user

        # Return generated response.
        return response

    def _get_login_user(self, user, *args, auto_login=True, user_permissions=None, user_groups=None, **kwargs):
        """Handles simulating user login with corresponding permissions/groups/etc.

        :param user: User to manipulate.
        :param auto_login: Bool indicating if User should be "logged in" to client or not.
        :param user_permissions: Django Permissions to give to User.
        :param user_groups: Django Groups to give to User.
        :return: Updated User object.
        """
        # Django imports here to avoid situational "Apps aren't loaded yet" error.
        from django.contrib.auth.models import AnonymousUser, Group, Permission

        # If not an anonymous user, attempt to get corresponding actual user object.
        if not isinstance(user, AnonymousUser):
            user = self.get_user(user)

        # Use anonymous user if either not set to auto-login, or provided user is inactive.
        if not auto_login or not hasattr(user, 'is_active') or not user.is_active:
            user = AnonymousUser()

        # If user is provided and is anonymous user object, then skip attempting any other login logic.
        if isinstance(user, AnonymousUser):
            return user

        # Handle mutable data defaults.
        user_permissions = user_permissions or []
        user_groups = user_groups or []

        # Handle possible types for Permissions.
        if isinstance(user_permissions, list) or isinstance(user_permissions, tuple):
            # Is array. This is expected.
            pass
        elif isinstance(user_permissions, str) or isinstance(user_permissions, Permission):
            # Is str or model instance. So assume single permission value.
            user_permissions = [user_permissions]
        else:
            # Invalid/unknown type. Raise error.
            raise TypeError('Provided Django Permissions must be either a str, array, or model format.')

        # Add all Permissions to provided user.
        for permission in user_permissions:
            user = self.add_user_permission(permission, user=user)

        # Handle possible types for Groups.
        if isinstance(user_groups, list) or isinstance(user_groups, tuple):
            # Is array. This is expected.
            pass
        elif isinstance(user_groups, str) or isinstance(user_groups, Group):
            # Is str or model instance. So assume single permission value.
            user_groups = [user_groups]
        else:
            # Invalid/unknown type. Raise error.
            raise TypeError('Provided Django Groups must be either a str, array, or model format.')

        # Add all Groups to provided user.
        for group in user_groups:
            user = self.add_user_group(group, user=user)

        # Optional hook to run additional authentication logic/setup on User.
        # For example, if project has 2-Factor setup that needs to be run.
        user = self._get_login_user__extra_user_auth_setup(
            user,
            *args,
            auto_login=True,
            user_permissions=None,
            user_groups=None,
            **kwargs,
        )

        # Ensure by this point that we have a proper instance of the User object.
        # Accounts for if wonky logic was used in extra_auth_setup, which is defined by the project importing
        # this package, so we have no control over.
        user = self.get_user(user)

        # Handle logging in with user.
        # This forces all response objects to act like this user is logged in for all page accesses.
        # Otherwise, it will act like an anonymous user is navigating the site.
        if auto_login:
            self.client.force_login(user)

        # Return modified user instance.
        return user

    def get_page_title(self, response):
        """Parses out title HTML element from provided response.

        :param response: Response object or response content to get title from.
        :return: Parsed out response title, formatted to have extra whitespace removed.
        """
        # Handle for provided response types.
        if isinstance(response, HttpResponseBase):
            response = response.content.decode('utf-8')
        elif isinstance(response, bytes):
            response = response.decode('utf-8')

        # Find title element.
        response_title = re.findall(r'<title(?:>| )([\S\s]+?)(?:</|</ |< /)title>', response)

        # Check that some value was found.
        # Certain response types may have no title, such as file download responses.
        if response_title is None:
            # No tags found. Convert to empty string.
            response_title = ''

        elif response_title is not None:
            # Tag was found.

            # Check how many title tags were found.
            if len(response_title) > 1:
                # Multiple headers were found. Raise error and direct user to helper h1 documentation.
                raise AssertionError(textwrap.dedent(
                    """
                    Found multiple titles ({0} total). There should only be one <title> tag per page.
                    For further reference on <title> tags, consider consulting:
                        * https://www.w3schools.com/tags/tag_title.asp
                        * https://developer.mozilla.org/en-US/docs/Web/HTML/Element/title
                    """.format(len(response_title))
                ).strip())

            elif len(response_title) == 0:
                # No title text was found. Return empty string.
                response_title = ''

            elif len(response_title) == 1:

                # Pull from capture group.
                response_title = response_title[0]

                # Strip any newlines, if present.
                response_title = re.sub(r'(\n|\r)+', '', response_title)

                # Remove any repeating whitespace, plus any outer whitespace.
                response_title = re.sub(r'(\s)+', ' ', response_title).strip()

        # Return formatted title value.
        return response_title

    def get_page_header(self, response):
        """Parses out H1 header HTML element from provided response.

        :param response: Response object or response content to get H1 header from.
        :return: Parsed out response header, formatted to have extra whitespace removed.
        """
        # Handle for provided response types.
        if isinstance(response, HttpResponseBase):
            response = response.content.decode('utf-8')
        elif isinstance(response, bytes):
            response = response.decode('utf-8')

        # Find header element.
        response_header = re.findall(r'<h1(?:>| )([\S\s]+?)(?:</|</ |< /)h1>', response)

        # Check that some value was found.
        # Handles if response did not have the H1 header element defined for some reason.
        # For example, likely to occur in responses that provide file downloads.
        if response_header is None:
            # No tags found. Convert to empty string.
            response_header = ''

        elif response_header is not None:
            # Tag was found.

            # Check how many header tags were found.
            if len(response_header) > 1:
                # Multiple headers were found. Raise error and direct user to helper h1 documentation.
                raise AssertionError(textwrap.dedent(
                    """
                    Found multiple headers ({0} total). There should only be one <h1> tag per page.
                    For further reference on <h1> tags, consider consulting:
                        * https://www.w3schools.com/tags/tag_hn.asp
                        * https://developer.mozilla.org/en-US/docs/Web/HTML/Element/Heading_Elements
                    """.format(len(response_header))
                ).strip())

            elif len(response_header) == 0:
                # No headers text was found. Return empty string.
                response_header = ''

            elif len(response_header) == 1:

                # Pull from capture group.
                response_header = response_header[0]

                # Strip any newlines, if present.
                response_header = re.sub(r'(\n|\r)+', '', response_header)

                # Remove any repeating whitespace, plus any outer whitespace.
                response_header = re.sub(r'(\s)+', ' ', response_header).strip()

            # Return formatted header value.
            return response_header

    def get_context_messages(self, response):
        """Parses out context messages from provided response.

        :param response: Response object or response context to get messages from.
        :return: Parsed out response messages.
        """
        # Handle for provided response types.
        if isinstance(response, HttpResponseBase):
            context = response.context
        else:
            context = response

        # Attempt to parse messages from context.
        found_messages = []
        if (context is not None) and ('messages' in context) and (len(context['messages']) > 0):
            # Messages found in response context.
            messages = context['messages']

            for message in messages:
                found_messages.append(str(message.message).strip())

        # Return found messages.
        return found_messages

    def _trim_response_content(self, response, content_starts_after=None, content_ends_before=None):
        """Trims response content, by trimming values at start or end of page content.

        :param response:
        :param content_starts_after:
        :param content_ends_before:
        :return: Dictionary of [sanitized content, trimmed/truncated content].
        """
        strip_err_msg = 'Could not find "{0}" value in content response. Provided value was:\n{1}'

        # Sanitize and format response content.
        minimized_content = self.get_minimized_response_content(response, strip_newlines=True)
        truncated_content = minimized_content

        # Rename variables for internal readability.
        strip_actual_start = content_starts_after
        strip_actual_end = content_ends_before

        if strip_actual_start:
            # Value passed that expected_content should occur AFTER.
            # Find first instance (from top of HTML output) of where this value occurs,
            # and then strip this and all above output.

            # First check that value actually exists in provided response.
            # Because we can't strip if this initial value is not present.
            stripped_start = self.get_minimized_response_content(strip_actual_start, strip_newlines=True)
            if stripped_start not in truncated_content:
                display_start = self.get_minimized_response_content(strip_actual_start, strip_newlines=False)
                self.fail(strip_err_msg.format('content_starts_after', display_start))
            # If we made it this far, then value was found. Remove.
            truncated_content = stripped_start.join(truncated_content.split(stripped_start)[1:])

        if strip_actual_end:
            # Value passed that expected_content should occur BEFORE.
            # Find first instance (from bottom of HTML output) of where this value occurs,
            # and then strip this and all below output.

            # First check that value actually exists in provided response.
            # Because we can't strip if this initial value is not present.
            stripped_end = self.get_minimized_response_content(strip_actual_end, strip_newlines=True)
            if stripped_end not in truncated_content:
                display_end = self.get_minimized_response_content(strip_actual_end, strip_newlines=False)
                self.fail(strip_err_msg.format('content_ends_before', display_end))
            # If we made it this far, then value was found. Remove.
            truncated_content = stripped_end.join(truncated_content.split(stripped_end)[:1])

        # Return both sanitized original content, and the stripped equivalent.
        return {
            'minimized_content': minimized_content,
            'truncated_content': truncated_content,
        }

    # endregion Helper Functions

    # region Hook Functions

    def _assertResponse__pre_builtin_tests(
        self,
        url, *args,
        get=True, data=None,
        expected_status=200,
        expected_redirect_url=None,
        url_args=None, url_kwargs=None, redirect_args=None, redirect_kwargs=None,
        expected_title=None, expected_header=None, expected_messages=None, expected_content=None,
        auto_login=True, user=None, user_permissions=None, user_groups=None,
        ignore_content_ordering=False, content_starts_after=None, content_ends_before=None,
        **kwargs,
    ):
        """Hook function to allow injecting code prior to running any of the built-in assertResponse() tests.

        For maximum usability, this function receives all args/kwargs provided to the default assertResponse()
        function.

        This hook does nothing by default, and is exclusively provided for custom extension logic.
        """
        pass

    def _assertResponse__post_builtin_tests(
        self,
        url, *args,
        get=True, data=None,
        expected_status=200,
        expected_redirect_url=None,
        url_args=None, url_kwargs=None, redirect_args=None, redirect_kwargs=None,
        expected_title=None, expected_header=None, expected_messages=None, expected_content=None,
        auto_login=True, user=None, user_permissions=None, user_groups=None,
        ignore_content_ordering=False, content_starts_after=None, content_ends_before=None,
        **kwargs,
    ):
        """Hook function to allow injecting code after running all of the built-in assertResponse() tests.

        For maximum usability, this function receives all args/kwargs provided to the default assertResponse()
        function.

        This hook does nothing by default, and is exclusively provided for custom extension logic.
        """
        pass

    def _get_login_user__extra_user_auth_setup(
        self,
        user,
        *args,
        auto_login=True,
        user_permissions=None, user_groups=None,
        **kwargs,
    ):
        """Hook function, to allow running extra authentication setup logic on User object.

        Useful such as for running things like 2-Factor setup logic for User.

        For maximum usability, this function receives all args/kwargs provided to the default _get_login_user()
        function.

        This hook does nothing by default, and is exclusively provided for custom extension logic.
        :return: Must return a User object. Either the original one as provided, or an updated version based on
                 custom logic.
        """
        return user

    # endregion Hook Functions


# Define acceptable imports on file.
__all__ = [
    'IntegrationTestCase',
]
